/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * options.h -- routines for parsing options
 */

#ifndef _MU_OPTIONS_H
#define _MU_OPTIONS_H

#include <stdio.h>
#include <dirent.h>

#include "compat.h"		/* For __attribute__() */

/* The maximum length of an error message. */
#define MU_OPT_ERR_MAX	255

#define MU_OPT_PERMUTE		0b00000001
#define MU_OPT_BUNDLE		0b00000010
#define MU_OPT_CONTINUE		0b00000100
#define MU_OPT_ALLOW_INVALID	0b00001000

#define MU_HELP_SHORT		0b00000001
#define MU_HELP_LONG		0b00000010
#define MU_HELP_MAN_SHORT	0b00000100
#define MU_HELP_MAN_LONG	0b00001000
#define MU_HELP_ENV		0b00010000
#define MU_HELP_BOTH		(MU_HELP_SHORT | MU_HELP_LONG)
#define MU_HELP_MAN_BOTH	(MU_HELP_MAN_SHORT | MU_HELP_MAN_LONG)
#define MU_HELP_ALL (MU_HELP_BOTH | MU_HELP_MAN_BOTH | MU_HELP_ENV)

/* Error codes. */
#define MU_OPT_ERR_PARSE	-1
#define MU_OPT_ERR_IO		-2
#define MU_OPT_ERR_CALLBACK 	-3

/* Check if a value is an error code. */
#define MU_OPT_ERR(value) ((value) < 0)

/* A convenience macro to be called from main(). */
#define MU_SHIFT_ARGS(amount) mu_shift_args(&argc, &argv, amount)

enum MU_OPT_HAS_ARG {
  MU_OPT_NONE = 0,
  MU_OPT_OPTIONAL,
  MU_OPT_REQUIRED
};

enum MU_OPT_ARG_TYPE {
  /* Start with 1 so that if this is not specified and it should be,
     an error will be raised rather than just defaulting to
     MU_OPT_BOOL. */
  MU_OPT_BOOL = 1,
  MU_OPT_INT,
  MU_OPT_FLOAT,
  MU_OPT_STRING,
  MU_OPT_FILE,
  MU_OPT_DIRECTORY,
  MU_OPT_SUBOPT
};

/* An opaque data structure to pass to mu_parse_opts(). */
typedef struct _mu_opt_data {
  char stop_parsing, skip_env;
} MU_OPT_DATA;

typedef struct _mu_opt {
  enum MU_OPT_HAS_ARG has_arg; /* Whether the option has an argument */
  enum MU_OPT_ARG_TYPE arg_type; /* The type of the argument */
  int *found_opt;	/* Whether or not we found this option */
  int *found_arg;	/* Whether or not we found an argument */
  union {
    struct {
      /* Regular option. */
      char short_opt;		/* The short option (if any) */
      const char *long_opt;	/* The long option (if any) */
    };
    /* Suboption. */
    const char *subopt_name;	/* The suboption name (if any) */
  };
  const char *env_var;	       /* The environment variable (if any) */
  union {
    struct {
      /* Simple arguments. */
      void *arg;	     /* Option argument (if any) */
      const char **argstr;   /* The string form of the argument */
      /* The default values for '*arg'. */
      union {
	int bool_default;
	long int_default;
	double float_default;
	const char *string_default;
	FILE *file_default;
	DIR *dir_default;
      };
      /* The callback should return zero on success or nonzero on
	 error, in which case the last argument (which must not be
	 written beyond OPT_ERR_MAX, including the terminating null
	 byte) must be set to a string describing the error. */
      union {
	int (*callback_none)(void *, char *);
	int (*callback_bool)(int, int, void *, char *);
	int (*callback_int)(int, long, void *, char *);
	int (*callback_float)(int, double, void *, char *);
	int (*callback_string)(int, const char *, void *, char *);
	int (*callback_file)(int, const char *, FILE *, void *, char *);
	int (*callback_directory)(int, const char *, DIR *, void *, char *);
      };
      void *cb_data;		/* Data to pass to that function */
      void (*cb_data_destructor)(void *); /* Free `cb_data' */
      union {
	struct {
	  long lower, upper; /* Lower and upper bounds for integers */
	} ibound;
	struct {
	  double lower, upper; /* Lower and upper bounds for floats */
	} fbound;
	const char *file_mode;	/* Mode to pass to fopen() */
      };
    };
    /* Suboption arguments. */
    struct _mu_opt *subopts;	/* The suboptions themselves */
  };
  const char *arg_help;	/* Format of the argument (e.g. "FILE", "WxH",
			   etc.) */
  const char *help;	/* Description of the option */
} MU_OPT;

int mu_parse_opts(int, char **, MU_OPT *, int, ...);
int mu_parse_opts_env(int, char **, char **, MU_OPT *, int, ...);
int mu_parse_subopts(const char *, char *, MU_OPT *);
void mu_shift_args(int *, char ***, int);
void mu_free_opts(MU_OPT *);
size_t mu_concat_opts(MU_OPT **, const MU_OPT *, const MU_OPT *, size_t);
void mu_create_help_option(MU_OPT *, const char *, const char *,
			   const char *, const char *, const char *,
			   const char *, const char *, const char *,
			   const char *, int);
int mu_format_help(FILE *, const char *, const char *,
		   const char *, const char *, const MU_OPT *);
char *mu_format_help_string(const char *, const char *, const char *,
			    const char *, const MU_OPT *,
			    unsigned short, unsigned short)
  __attribute__((malloc));
int mu_format_help_man(FILE *, const char *, const char *,
		       const char *, const char *, const char *,
		       const char *, const char *, const char *,
		       const char *, const MU_OPT *);
char *mu_format_help_man_string(const char *, const char *,
				const char *, const char *,
				const char *, const char *,
				const char *, const char *,
				const char *, const MU_OPT *)
  __attribute__((malloc));

#endif /* !_MU_OPTIONS_H */
