/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * test.c -- common functions that can be used by all test programs
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

/* Mu includes */
#include <options.h>
#include <format.h>
#include <safe.h>

#include "test.h"

/*
 * Print the name of the program (passed in `data') and exit.
 */
int _test_print_name(void *data, char *err) {
  const char *name = data;
  if (puts(name) == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot print program name, '%s': %s",
	     name, strerror(errno));
    return 1;
  }
  exit(0);
}

/*
 * Get an integer and a string from `str'. Returns zero on success or
 * nonzero on error, in which case an error message will be copied to
 * `err', and '*num' and '*outstr' will be left unchanged.
 */
int get_int_str(int *num, const char **outstr,
		const char *str, char *err) {
  long result;
  char *endptr;

  errno = 0;
  result = strtol(str, &endptr, 0);
  if (result < INT_MIN || result > INT_MAX)
    errno = ERANGE;
  if (errno) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': %s", str, strerror(errno));
    return 1;
  }
  if (!*endptr) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "'%s': missing second argument", str);
    return 1;
  }

  while (*endptr && isspace(*endptr))
    endptr++;
  if (*endptr != ',') {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': not an integer", str);
    return 1;
  }
  endptr++;

  *num = result;
  *outstr = endptr;
  return 0;
}

/*
 * Get a string and a single character from `str'. The string,
 * returned in '*outstr', will be dynamically allocated and must be
 * freed. Returns zero on success or nonzero on error, in which case
 * an error message will be copied to `err', and '*outstr' and '*c'
 * will be left unchanged.
 */
int get_str_char(char **outstr, char *c, const char *str, char *err) {
  const char *comma;

  comma = strchr(str, ',');
  if (!comma) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "'%s': missing second argument", str);
    return 1;
  }
  if (!comma[1] || comma[2]) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "'%s': must be a single character", comma + 1);
    return 1;
  }

  *outstr = mu_xstrndup(str, comma - str);
  *c = comma[1];
  return 0;
}

/*
 * Parse a string into one or two `size_t's. Returns zero on success
 * or nonzero on error, in which case an error message will be copied
 * to `err' and '*size{,1,2}' will be unchanged.
 */

int get_size_1(size_t *size, const char *str, char *err) {
  ullong_t result;
  char *endptr;

  errno = 0;
  result = strtoullt(str, &endptr, 0);
  if (errno) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': %s", str, strerror(errno));
    return 1;
  }
  if (!*str || *endptr) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': not an integer", str);
    return 1;
  }

  *size = result;
  return 0;
}

int get_size_2(size_t *size1, size_t *size2,
	       const char *str, char *err) {
  ullong_t result1, result2;
  char *endptr1, *endptr2;

  if (!*str) {
    strncpy(err, "'': not an integer", MU_OPT_ERR_MAX);
    return 1;
  }
  errno = 0;
  result1 = strtoullt(str, &endptr1, 0);
  if (errno) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': %s", str, strerror(errno));
    return 1;
  }
  if (!*endptr1) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "'%s': missing second argument", str);
    return 1;
  }

  while (*endptr1 && isspace(*endptr1))
    endptr1++;
  if (*endptr1 != ',') {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': not an integer", str);
    return 1;
  }
  endptr1++;

  errno = 0;
  result2 = strtoullt(endptr1, &endptr2, 0);
  if (errno) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': %s",
	     endptr1, strerror(errno));
    return 1;
  }
  if (!*endptr1 || *endptr2) {
    snprintf(err, MU_OPT_ERR_MAX, "'%s': not an integer", endptr1);
    return 1;
  }

  *size1 = result1;
  *size2 = result2;
  return 0;
}
