/********************************************************************\
 * Mu (Miscellaneous Utilities) is a general convenience library.   *
 * Copyright (C) 2020  Asher Gordon <AsDaGo@posteo.net>             *
 *                                                                  *
 * This file is part of Mu.                                         *
 *                                                                  *
 * Mu is free software: you can redistribute it and/or modify it    *
 * under the terms of the GNU General Public License as published   *
 * by the Free Software Foundation, either version 3 of the         *
 * License, or (at your option) any later version.                  *
 *                                                                  *
 * Mu is distributed in the hope that it will be useful, but        *
 * WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public        *
 * License along with Mu.  If not, see                              *
 * <https://www.gnu.org/licenses/>.                                 *
\********************************************************************/

/*
 * test.h -- common header to be included by all test programs
 */

#ifndef _TEST_H
#define _TEST_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

/* Mu includes */
#include <options.h>
#include <format.h>

#define STR(x)	#x
#define XSTR(x)	STR(x)

/* Define strtoullt() to the next best thing if we don't have
   strtoull(). `ullong_t' is the type that strtoullt() returns, and
   ULLONG_T_MAX is the maximum value a `ullong_t' can hold. */
#ifdef HAVE_STRTOULL
typedef unsigned long long ullong_t;
# define ULLONG_T_MAX ULLONG_MAX
# define PRIdULLT "llu"
# define strtoullt strtoull
#else
typedef unsigned long ullong_t;
# define PRIdULLT "lu"
# ifdef HAVE_STRTOUL
#  define ULLONG_T_MAX ULONG_MAX
#  define strtoullt strtoul
# else
#  define ULLONG_T_MAX LONG_MAX
#  define strtoullt(n, e, b) (ullong_t)strtol(n, e, b)
# endif
#endif

/* This must be the last "option" to appear in the option list. */
#define TEST_OPT_END						\
  _test_name_opt, _test_version_opt, { 0 }, { 0 }, { 0 }, { 0 }

#define _TEST_DESC_START				\
  "Test " TEST_SUBJECT ". This program is part of the "	\
  PACKAGE_STRING " test suite."

#ifdef TEST_DESC
# define _TEST_FULL_DESC _TEST_DESC_START "\n\n" TEST_DESC
#else
# define _TEST_FULL_DESC _TEST_DESC_START
#endif

#define _TEST_NOTES						\
  "Report bugs and test failures to " PACKAGE_BUGREPORT "."

#ifndef TEST_USAGE
# define TEST_USAGE NULL
#endif

/* A convenience macro to create a help option. */
#define test_create_help_option(options)			\
  mu_create_help_option(options, argv[0], TEST_USAGE, "test "	\
			TEST_SUBJECT, _TEST_FULL_DESC,		\
			_TEST_NOTES, "1", NULL,			\
			PACKAGE_STRING " Test Suite",		\
			NULL, MU_HELP_ALL)

/* Like the above, but format the help directly instead of creating an
   option for it. */
#define test_format_help(stream, options)		\
  mu_format_help(stream, argv[0], TEST_USAGE,		\
		 _TEST_FULL_DESC, _TEST_NOTES, options)

/*
 * An option to print out the name of the program (argv[0]). We have
 * to #define it so we can use 'argv[0]'.
 */
#define _test_name_opt {					\
    .long_opt	= "name",					\
    .has_arg	= MU_OPT_NONE,					\
    .callback_none = _test_print_name,				\
    .cb_data	= argv[0],					\
    .help	= "print the name of this program and exit"	\
  }

#ifdef TEST_NAME
/*
 * Print version information and exit.
 */
static int _test_print_version(void *data, char *err) {
  unsigned short col = 0;
  if (mu_format(stdout, &col, 0, 70, 0, 0,
TEST_NAME ": part of the " PACKAGE_STRING " test suite.\n\
Copyright (C) 2020 Asher Gordon <AsDaGo@posteo.net>\n\
\n\
This test suite is part of " PACKAGE_NAME ".\n\
\n"
PACKAGE_NAME " comes with ABSOLUTELY NO WARRANTY. You can \
redistribute and/or modify " PACKAGE_NAME " under the terms of the \
GNU General Public License, version 3 or, at your option, any later \
version.\n\
\n\
For more information, see the file COPYING.\n") < 0 ||
      fflush(stdout) == EOF) {
    snprintf(err, MU_OPT_ERR_MAX,
	     "cannot print version information: %s", strerror(errno));
    return 1;
  }
  exit(0);
}

/*
 * An option to print version information. This is used by
 * ../testsuite.
 */
static const MU_OPT _test_version_opt = {
  .short_opt	= 'v',
  .long_opt	= "version",
  .has_arg	= MU_OPT_NONE,
  .callback_none = _test_print_version,
  .help		= "print version information and exit"
};
#endif /* TEST_NAME */

int _test_print_name(void *, char *);
int get_int_str(int *, const char **, const char *, char *);
int get_str_char(char **, char *, const char *, char *);
int get_size_1(size_t *, const char *, char *);
int get_size_2(size_t *, size_t *, const char *, char *);

#endif /* !_TEST_H */
